# cython: language_level=3, boundscheck=False

from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp cimport bool

from ld_estimator.linkage import LD

cdef extern from 'haps.h' namespace 'ld_estimator':
  cdef cppclass Haps[T]:
      T aa
      T ab
      T ba
      T bb

cdef extern from 'linkage.h' namespace 'ld_estimator':
  cdef struct Linkage:
      double dprime
      double loglikelihood
      double r_squared
      double ci_low
      double ci_high
      Haps[double] freqs
      vector[string] phase

cdef extern from 'ld.h' namespace 'ld_estimator':
    Linkage pairwise(vector[vector[string]], vector[vector[string]], vector[bool]) except +

cdef to_bytes(var):
    return [[str(a).encode('utf8') for a in g] for g in var]

def pairwise_ld(var1, var2, vector[bool] ploidy):
    # only convert to allele vectors to bytes if not done already
    if not isinstance(var1[0][0], bytes):
        var1 = to_bytes(var1)
    if not isinstance(var2[0][0], bytes):
        var2 = to_bytes(var2)
    
    if len(var1) != len(var2):
      raise ValueError('genotypes lists need to be the same length')
    if len(var1) == 0:
      raise ValueError('zero genotypes in lists supplied for LD')
    
    try:
        ld = pairwise(var1, var2, ploidy)
    except ValueError:
        return None
    
    return LD(ld.dprime, ld.loglikelihood, ld.r_squared, ld.ci_low, ld.ci_high,
      [ld.freqs.aa, ld.freqs.ab, ld.freqs.ba, ld.freqs.bb], [x.decode('utf8') for x in ld.phase])

cdef extern from 'utils.h' namespace 'ld_estimator':
    bool is_monomorphic(vector[vector[string]]) except +

def _is_monomorphic(var):
  return is_monomorphic(to_bytes(var))



''' class to store LD results
'''

class LD(object):
    ''' simple class to store linkage disequilibrium results
    '''
    def __init__(self, dprime, loglikelihood, r_squared, ci_low, ci_high, freqs, phase):
        self.dprime = dprime
        self.loglikelihood = loglikelihood
        self.r_squared = r_squared
        self.ci_low = ci_low
        self.ci_high = ci_high
        self.freqs = freqs
        self.phase = phase

    def __repr__(self):
        return f'LD({self.dprime}, {self.loglikelihood}, {self.r_squared}, ' \
            f'{self.ci_low}, {self.ci_high}, {self.freqs}, {self.phase})'



''' provides function to calculate LD within a genome region, given a VCF
'''

from itertools import combinations

from ld_estimator.utils import prepare_vcf, check_ld
from ld_estimator.pairwise import _is_monomorphic

def region_ld(vcf, chrom, start, end, subset=None, sexes=None, build='grch37'):
    ''' calculate LD within a genome region

    Args:
        vcf: path to vcf, or pysam.VariantFile object
        chrom: chromosome.
        start: start of region to calculate LD within
        end: end of region to calculate LD within
        subset: list of sample IDs to include in LD calculation. Default is to
            include every sample.
        sexes: list of sexes of samples (matching subset order). Only needed
            if variant on a sex chromosome and not in a pseudoautosomal region.
        build: genome build to use (defaults to 'grch37')

    Returns:
        list of LD estimates, each as a dict with [chrom, var1_pos, var2_pos,
            r_squared and dprime] keys.
    '''

    vcf, ploidy = prepare_vcf(vcf, chrom, start, subset, sexes, build)

    lds = []

    variants = []
    for var in vcf.fetch(chrom, start, end):
        geno = [var.samples[x].alleles for x in var.samples]
        if _is_monomorphic(geno):
            continue
        geno = [tuple(map(str.encode, x)) for x in geno]
        variants.append((var, geno))

    for (var1, var1_geno), (var2, var2_geno) in combinations(variants, 2):
        data = check_ld(var1, var1_geno, var2, var2_geno, ploidy)
        lds.append(data)

    return lds



''' unit tests for LD class
'''

import unittest

from ld_estimator import pairwise_ld

class TestLDEstimator(unittest.TestCase):
    ''' class to unit test LD calculations
    '''
    def test_pairwise_ld(self):
        ''' check pairwise_ld works correctly
        '''
        var1 = [('A', 'A'), ('A', 'A'), ('A', 'G'), ('G', 'A'), ('G', 'G'),
                ('A', 'G'), ('A', 'A'), ('A', 'A'), ('G', 'G')]
        var2 = [('A', 'A'), ('A', 'A'), ('A', 'G'), ('G', 'A'), ('G', 'G'),
                ('A', 'G'), ('A', 'A'), ('G', 'G'), ('G', 'G')]
        is_haploid = [False, False, False, False, False, False, False, True, True, True]

        linkage = pairwise_ld(var1, var2, is_haploid)
        self.assertEqual(linkage.dprime, 0.9999999995259259)
        self.assertEqual(linkage.r_squared, 0.7714285713447618)

    def test_pairwise_ld_low_dprime(self):
        ''' check pairwise_ld works correctly for variants with low D'
        '''
        var1 = [('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'),
                ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'),
                ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'),
                ('G', 'G'), ('G', 'A'), ('G', 'G'), ('G', 'A'), ('G', 'G'),
                ('G', 'G'), ('G', 'G'), ('G', 'G')]

        var2 = [('G', 'G'), ('G', 'G'), ('A', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'),
                ('G', 'A'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('A', 'G'), ('G', 'G'),
                ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('A', 'G'), ('G', 'G'),
                ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'A')]

        is_haploid = [False] * len(var1)
        linkage = pairwise_ld(var1, var2, is_haploid)
        self.assertEqual(linkage.dprime, 0.37448437109790317)
        self.assertEqual(linkage.r_squared, 0.05227073010963888)

    def test_pairwise_ld_haploid(self):
        ''' check pairwise_ld can handle single entry genotypes
        '''
        var1 = [('G'), ('G'), ('A'), ('G', 'G'), ('G', 'A'),
            ('G', 'G'), ('G', 'A'), ('G', 'G'), ('G', 'G'), ('G'), ('G', 'G')]
        
        var2 = [('G'), ('G'), ('A'), ('G', 'G'), ('G', 'G'),
            ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G'), ('G', 'A')]
        
        is_haploid = [False] * len(var1)
        linkage = pairwise_ld(var1, var2, is_haploid)
        self.assertEqual(linkage.dprime, 0.5925925925925927)
        self.assertEqual(linkage.r_squared, 0.24951267056530224)

    def test_pairwise_ld_monomorphic(self):
        ''' test we don't compute LD with monomorphic markers
        '''
        var1 = [('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G')]
        var2 = [('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G')]
        is_haploid = [False] * len(var1)
        self.assertIsNone(pairwise_ld(var1, var2, is_haploid))

        # check it gives none even if only one marker is monomorphic
        var3 = [('G', 'A'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G'), ('G', 'G')]
        self.assertIsNone(pairwise_ld(var1, var3, is_haploid))

    def test_lacks_haplotypes(self):
        ''' test pairwise_ld when missing some haplotypes
        '''
        var1 = [('0', '0'), ('0', '0'), ('0', '0'), ('1', '1'), (None, None)]
        var2 = [('0', '0'), ('0', '0'), ('0', '0'), (None, None), ('1', '1')]
        is_haploid = [False] * len(var1)
        linkage = pairwise_ld(var1, var2, is_haploid)
        self.assertEqual(linkage.dprime, 1)
        self.assertEqual(linkage.r_squared, 0)
        self.assertEqual(linkage.freqs, [0, 0, 0, 0])



''' package to calculate linkage disequilibrium
'''
from pkg_resources import get_distribution

from ld_estimator.pairwise import pairwise_ld
from ld_estimator.region import region_ld
from ld_estimator.site import site_ld

__version__ = get_distribution('ld_estimator').version



''' provides function to calculate LD against a single site in a VCF
'''

from ld_estimator.utils import prepare_vcf, check_ld

def get_intervals(positions, window):
    ''' find windows of genomic regions for getting variants within.

    This merges overlapping intervals
    '''
    intervals = []
    for pos in sorted(positions):
        if intervals and pos - window < intervals[-1][1]:
            intervals[-1][1] = pos + window
        else:
            intervals.append([pos - window, pos + window])

    return intervals

def site_ld(vcf, chrom, positions, window=100000, subset=None, sexes=None, build='grch37'):
    ''' calculate LD at specified site/s to all surrounding variants

    Args:
        vcf: path to vcf, or pysam.VariantFile object
        chrom: chromosome.
        positions: position/s of variant/s to calculate LD against.
        window: number of base pairs to check LD within
        subset: list of sample IDs to include in LD calculation.
            Default is to include every sample.
        sexes: list of sexes of samples (matching subset order). Only needed
            if variant on a sex chromosome and not in a pseudoautosomal region.
        build: genome build to use (defaults to 'grch37')

    Returns:
        list of LD estimates, each as a dict with [chrom, var1_pos, var2_pos,
            var1_id, var2_id, r_squared, dprime and phase] keys.
    '''
    if not isinstance(positions, list):
        positions = [positions]

    vcf, ploidy = prepare_vcf(vcf, chrom, positions[0], subset, sexes, build)

    variants = []
    for pos in positions:
        entries = list(x for x in vcf.fetch(chrom, pos-1, pos) if x.pos == pos)
        if entries == []:
            raise ValueError(f'no variant found at: {chrom}:{pos}')
        var = entries[0]
        geno = [var.samples[x].alleles for x in var.samples]
        try:
            geno = [tuple(map(str.encode, x)) for x in geno]
        except TypeError:
            # TODO: fix for missing genotypes!
            continue
        variants.append([var, geno])

    lds = []
    for (start, end) in get_intervals(positions, window):
        for var2 in vcf.fetch(chrom, start, end):
            var2_geno = [var2.samples[x].alleles for x in var2.samples]
            for var1, var1_geno in variants:
                if var1 == var2:
                    continue

                data = check_ld(var1, var1_geno, var2, var2_geno, ploidy)
                lds.append(data)

    return lds



''' setup.py file for ld_estimator package
'''

import io
import sys
from setuptools import setup, Extension
from Cython.Build import cythonize

EXTRA_COMPILE_ARGS = ['-std=c++11']
EXTRA_LINK_ARGS = []
if sys.platform == "darwin":
    EXTRA_COMPILE_ARGS += ["-stdlib=libc++", "-mmacosx-version-min=10.9"]
    EXTRA_LINK_ARGS += ["-stdlib=libc++", "-mmacosx-version-min=10.9"]

EXT = cythonize([
    Extension('ld_estimator.pairwise',
              extra_compile_args=EXTRA_COMPILE_ARGS,
              extra_link_args=EXTRA_LINK_ARGS,
              sources=['ld_estimator/pairwise.pyx',
                       'src/ld.cpp',
                       'src/af.cpp',
                       'src/confidence.cpp',
                       'src/dprime.cpp',
                       'src/flip.cpp',
                       'src/frequencies.cpp',
                       'src/likelihoods.cpp',
                       'src/optimize.cpp',
                       'src/tallies.cpp',
                       'src/utils.cpp',
                       ],
              include_dirs=['src/'],
              language='c++')
    ])

setup(name="ld_estimator",
      description='Package for estimating linkage disequilibrium',
      long_description=io.open('README.md', encoding='utf-8').read(),
      long_description_content_type='text/markdown',
      version="1.0.7",
      author="Jeremy McRae",
      author_email="jmcrae@illumina.com",
      url='https://github.com/jeremymcrae/ld_estimator',
      license='MIT',
      packages=['ld_estimator'],
      install_requires=['pysam'],
      python_requires='>3.6.0',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
      ],
      ext_modules=EXT,
      test_suite='tests')



''' some common functions
'''

import pysam

from ld_estimator import pairwise_ld

PSEUDOAUTOSOMAL = {'grch37': {'X': [(60001, 2699520), (154931044, 155260560)],
                              'Y': [(10001, 2649520), (59034050, 59363566)]},
                   'grch38': {'X': [(10001, 2781479), (155701383, 156030895)],
                              'Y': [(10001, 2781479), (56887903, 57217415)]}}
MALE_CODES = set(['m', 'male', '1'])
FEMALE_CODES = set(['f', 'female', '2'])

def prepare_vcf(vcf, chrom, pos, subset, sexes, build='grch37'):
    ''' get a pysam.VariantFile ready for calculating LD and the ploidy state

    Opens the VCF file, and selects only the required samples (all samples by
    default).

    Args:
        vcf: path to VCF (or pysam.VariantFile handle)
        chrom: chromosome of interest
        pos: nucleotide position to check ploidy at
        subset: list of sample IDs to subset on (or None)
        sexes: list of sample sexes (or None). Has to match the subset order if
            that is present, otherwise the VCF samples order
        build: genome build to use (defaults to 'grch37')

    Returns:
        tuple of (pysam.VariantFile, ploidy list)
    '''
    try:
        vcf = pysam.VariantFile(vcf)
    except AttributeError:
        pass

    if not is_autosomal(chrom, pos) and sexes is None:
        raise ValueError('Sample sexes required for non-autosomal sites: {chrom}:{pos}')

    if sexes is None and subset is None:
        sexes = ['female'] * len(vcf.header.samples)

    if subset is not None:
        all_ids = set(vcf.header.samples)
        masked = [x in all_ids for x in subset]
        sample_ids = [sample for x, sample in zip(masked, subset) if x]
        sexes = [sample for x, sample in zip(masked, sexes) if x]
        try:
            vcf.subset_samples(sample_ids)
        except ValueError:
            # can only subset once, check the sets match if done previously
            assert all_ids == set(sample_ids)
    else:
        sample_ids = list(vcf.header.samples)

    # re-sort the sexes based on the VCF sample order, so that the ploidy list
    # later matches the genotypes order
    indexes = dict(zip(sample_ids, range(len(subset))))
    sample_ids = [sample_ids[indexes[x]] for x in vcf.header.samples]
    sexes = [sexes[indexes[x]] for x in vcf.header.samples]

    return vcf, get_ploidy(chrom, pos, sample_ids, sexes, build)

def check_ld(var1, var1_geno, var2, var2_geno, ploidy):
    ''' get a dictionary with variant coords and LD details
    '''
    data = {'chrom': var1.chrom,
            'var1': {'pos': var1.pos, 'id': var1.id, 'ref': var1.ref, 'alt': var1.alts[0]},
            'var2': {'pos': var2.pos, 'id': var2.id, 'ref': var2.ref, 'alt': var2.alts[0]},
            'r_squared': None, 'dprime': None, 'phase': None}

    linkage = pairwise_ld(var1_geno, var2_geno, ploidy)
    if linkage is not None:
        data['r_squared'] = linkage.r_squared
        data['dprime'] = linkage.dprime
        data['phase'] = linkage.phase

    return data

def is_autosomal(chrom, pos, build='grch37', par_regions=None):
    ''' identify whether a genomic site is autosomal

    This includes pseudoautosomal regions as autosomal.

    Args:
        chrom: chromosome of variant
        pos: nucleotide position of variant
        build: genome build (defaults to 'grch37', but can also use 'grch38',
            'hg19', 'hg38' and use the default pseudoautosomal regions. If a
            a different genome build is needed, supply a own par_regions
            dictionary.
        par_regions: dict of pseudoautosomal regions, indexed by genome build,
            then chromosome. Regions are given as list of (start, end) tuples.
            For example {'grch37': {'X': [(1, 10), ...], 'Y': [(30, 40), ...]}}
    '''
    # standardise the genome build name
    build = build.lower()
    names = {'hg19': 'grch37', 'hg38': 'grch38'}
    build = names[build] if build in names else build

    # load the default PAR dict if none provided
    if not par_regions:
        par_regions = PSEUDOAUTOSOMAL

    # select the pseudoautosomal regions on the given chromosome
    chrom = chrom.strip('chr')
    par = par_regions[build]
    autosomal = chrom not in par or any(x[0] <= int(pos) <= x[1] for x in par[chrom])
    return autosomal

def get_ploidy(chrom, pos, samples, sexes, build='grch37'):
    ''' figure out the expected ploidy for each individual

    Given a list of sexes, and the chromosome and position, figure out whether
    the individuals are haploid (males on chrX). This ignores chrY, but that is
    ok, since chrY is haploid, and therefore LD is simple to assess.

    Currently this only checks ploidy at one site, but technically we should
    figure out ploidy at both sites being assessed for LD, but that gets tricky.
    '''
    if is_autosomal(chrom, pos, build):
        return [False] * len(samples)

    for sample, sex in zip(samples, sexes):
        sex = sex.lower()
        assert sex in MALE_CODES or sex in FEMALE_CODES, f'unknown sex ({sex}) in {sample}'

    return [sex.lower() in MALE_CODES for sex in sexes]
